#include "afflib.h"
#include "afflib_i.h"


/* evf file implementation */
struct evf_private {
    /* private data */
    int cur_seg;			// current metadata segment, for iterator
};

static inline struct evf_private *EVF_PRIVATE(AFFILE *af)
{
    return (struct evf_private *)(af->vnodeprivate);
}

/* Return 1 if a file is a evf file... */
static int evf_identify_file(const char *filename)
{
    return access(filename,R_OK)==0;	// if we can read it, it's evf...
}


/* Return the size of the evf file */
static int64 evf_filesize(AFFILE *af)
{
    struct stat sb;
    if(fstat(fileno(af->evf),&sb)){
	warn("evf_open: stat(%s): ",af->fname);
	return -1;			// kind of odd...
    }
    return sb.st_size;
}

static int evf_open(AFFILE *af)
{
    /* Evf is the passthrough system.
     * Right now, it is read only...
     */
    af->evf=fopen(af->fname,"rb");
    af->imagesize = evf_filesize(af);
    af->image_pagesize = EVF_PAGESIZE;
    af->vnodeprivate = (void *)calloc(1,sizeof(struct evf_private));
    struct evf_private *rp = EVF_PRIVATE(af);
    rp->cur_page = 0;

    return 0;
}

static int evf_close(AFFILE *af)
{
    struct evf_private *rp = EVF_PRIVATE(af);

    if(af->evf_popen){
	pclose(af->evf);
    }
    else {
	fclose(af->evf);
    }
    memset(rp,0,sizeof(*rp));		// clean object reuse
    free(rp);				// won't need it again
    return 0;
}

static int evf_get_seg(AFFILE *af,const char *name,unsigned long *arg,unsigned char *data,size_t *datalen)
{
    /* Right now, this only supports page segments */
    unsigned int segnum=0;
    if(sscanf(name,AF_PAGE,&segnum)!=1){
	return -1;			// unknown segment number
    }
    fflush(af->evf);			// make sure that any buffers are flushed

    int64 pos = (int64)segnum * af->image_pagesize; // where we are to start reading
    int64 bytes_left = af->imagesize - pos;	// how many bytes left in the file

    int bytes_to_read = af->image_pagesize; // copy this many bytes, unless
    if(bytes_to_read > bytes_left) bytes_to_read = bytes_left; // only this much is left
    
    if(arg) *arg = 0;			// arg is always 0
    if(datalen){
	if(*datalen==0 && data==0){ // asked for 0 bytes, so give the actual size
	    *datalen = bytes_to_read;
	    return 0;
	}
	if(*datalen < (unsigned)bytes_to_read){
	    *datalen = bytes_to_read;
	    return AF_ERROR_DATASMALL;
	}
    }
    if(data){
	/* TK: Get the segment number segnum (or offset 'pos')
	 * put the results in data.
	 * Return -1 if failure, 0 if success
	 */
	return -1;			// some kind of EOF?
    }
    return 0;				// no problems!
}


int evf_update_seg(AFFILE *af, const char *name,
		    unsigned long arg,const void *value,unsigned int vallen,int append)
{
    /* Write back to the EVF file. Currently unsupported. */
    return -1;				// some kind of error...
}


static int evf_stat(AFFILE *af,struct af_vnode_info *vni)
{
    vni->imagesize = -1;		// set to number of bytes in the image
    vni->pagesize  = EVF_PAGESIZE;	// set to the page size
    return 0;
}

static int evf_rewind_seg(AFFILE *af)
{
    struct evf_private *rp = EVF_PRIVATE(af);
    
    rp->cur_seg = 0;
    return 0;
}


static int evf_get_next_seg(AFFILE *af,char *segname,size_t segname_len,unsigned long *arg,
			unsigned char *data,size_t *datalen_)
{
    return -1;				// not implemented yet;
}


struct af_vnode vnode_evf = {
    AF_IDENTIFY_EVF,
    evf_identify_file,
    evf_open,
    evf_close,
    evf_stat,
    evf_get_seg,			// get seg
    evf_get_next_seg,			// get_next_seg
    evf_rewind_seg,			// rewind_seg
    evf_update_seg,			// update_seg
    0					// del_seg
};

