/*
 * acompare.cpp:
 *
 * Compare the contents of an ISO file to an AFF file.
 * Optionally, if they are equal, delete the ISO file
 */

/*
 * Copyright (c) 2005
 *	Simson L. Garfinkel and Basis Technology, Inc. 
 *      All rights reserved.
 *
 * This code is derrived from software contributed by
 * Simson L. Garfinkel
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Simson L. Garfinkel
 *    and Basis Technology Corp.
 * 4. Neither the name of Simson Garfinkel, Basis Technology, or other
 *    contributors to this program may be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY SIMSON GARFINKEL, BASIS TECHNOLOGY,
 * AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SIMSON GARFINKEL, BAIS TECHNOLOGy,
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.  
 */


#include "afflib.h"
#include "afflib_i.h"

#include <ctype.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <err.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <limits.h>
#include <string.h>
#include <zlib.h>
#include <openssl/md5.h>
#include <openssl/sha.h>
#include <assert.h>

#include <algorithm>
#include <cstdlib>
#include <vector>
#include <string>

using namespace std;

char *progname = "acompare";

int opt_remove = 0;
int opt_quiet  = 0;
int opt_batch  = 0;
int opt_all    = 0;
int opt_All    = 0;
int opt_page   = -1;
char *batch_ext = "";

void usage()
{
    printf("usage: %s [options] file1 file2...\n",progname);
    printf("options:\n");
    printf("       -b ext    --- Base. Use extension 'ext' for raw files\n");
    printf("       -r        --- Remove. If two files are equal, remove the ISO\n");
    printf("       -q        --- Quiet. No output except for errors\n");
    printf("       -a        --- print the numbers of differing sectors\n");
    printf("       -A        --- print the contents of differing sectors\n");
    //printf("       -n ###    --- print the contents of sector ### if it is different\n");
    printf("       -p ###    --- Just examine the differences on page ###\n");
    printf("\n");
    printf("Examples: For comparing ISO and AFF files:\n");
    printf("  %s -b img file.aff               --- compare file.aff and file.img\n",
	   progname);
    printf("  %s -b img file1.aff file2.aff... --- compare file1.aff, file1.img, etc.\n",
	   progname);
    printf("  %s -r Remove the ISO file if the two are equal.\n",progname);
    printf("  %s -q Quiet; no output except for errors.\n",progname);
    printf("\n");
    printf("Examples For comparing two AFF files:\n");
    printf("  %s  file1.aff file2.aff          --- compares file1 & file2\n",progname);
    exit(0);
}

void print_sector(unsigned char buf[512])
{
    for(int i=0;i<512;i++){
	if(isprint(buf[i])){
	    putchar(buf[i]);
	}
	else {
	    putchar('.');
	}
	if(i%64==63) putchar('\n');
    }
}


void compare_aff_iso(const char *aff_file,const char *iso_file)
{
     /* make sure the extension is right */
     char *cc = strrchr(aff_file,'.');
     if(!cc || strcmp(cc,".aff")!=0){
	 fprintf(stderr,"%s is not a valid AFF file.\n",aff_file);
	 exit(1);
     }

     /* See if they are the same file. That would be senseless */
     struct stat sa,si;

     if(stat(aff_file,&sa)) err(1,aff_file);
     if(stat(iso_file,&si)) err(1,iso_file);

     if(sa.st_dev==si.st_dev &&
	sa.st_ino==si.st_ino){
	 fprintf(stderr,"%s and %s are the same file!\n",aff_file,iso_file);
	 exit(1);
     }

     /* Kind of a hack. We use the af routines to open both files, since
      * they automatically fall-back to ISO files for non-AFF files.
      * In the process, we get to double-check the AFF routines as well...
      * (Don't worry; we'll also check the MD5.)
      */
     AFFILE *file1 = af_open(aff_file,O_RDONLY,0);
     if(!file1) err(1,"afopen %s",aff_file);
     
     AFFILE *file2 = af_open(iso_file,O_RDONLY,0);
     if(!file2) err(1,"afopen %s",iso_file);

     if(!opt_quiet){
	 printf("comparing %s and %s...\n",aff_file,iso_file);
     }

     /* Compare 1MB at a time, to minimize on seeks... */

     long long total_bytes = 0;
     struct timeval start;
     gettimeofday(&start,0);

     while(!af_eof(file1)){
	 char buf1[65536 * 16];
	 char buf2[65536 * 16];
	 size_t bytes1,bytes2;

	 if(!opt_quiet){
	     /* Calculate the elapsed time and print status information */
	     struct timeval now;
	    
	     gettimeofday(&now,0);
	     float percent_done = total_bytes*100.0/si.st_size;

	     printf("\r%qd out of %qd bytes compared  (%5.2f%%) ",
		    total_bytes,(long long)si.st_size,percent_done);

	     /* If we have non-zero time and non-zero bytes, report throughput */
	     long total_msec = (now.tv_sec-start.tv_sec)*1000 + (now.tv_usec-start.tv_usec)/1000;
	     long total_sec  = total_msec / 1000;

	     if(total_msec>0 && total_bytes>0){
		 printf("%6.2f MBytes/sec ",total_bytes/(total_msec * 1000.0));
	     }

	     /* Figure out how much time is left... */
	     if(percent_done>0){
		 long est_total_time = (long)(total_sec / (percent_done/100.0));
		 long est_time_remaining = est_total_time - total_sec;
		 printf("%2ld:%02ld ",est_time_remaining / 60, est_time_remaining % 60);
	     }
	     fflush(stdout);
	 }

	 bytes1 = af_read(file1,(unsigned char *)buf1,sizeof(buf1));
	 bytes2 = af_read(file2,(unsigned char *)buf2,sizeof(buf2));

	 if(bytes1 != bytes2){
	     fprintf(stderr,"\n");
	     fprintf(stderr,"Error at offset %qd. Attempted to read %lu bytes.\n",
		     total_bytes,(long unsigned int)sizeof(buf1));
	     fprintf(stderr,"afread(%s) = %lu bytes\n",aff_file,(long unsigned int)bytes1);
	     fprintf(stderr,"afread(%s) = %lu bytes\n",iso_file,(long unsigned int)bytes2);
	     exit(2);
	 }

	 if(memcmp(buf1,buf2,bytes1)!=0){
	     fprintf(stderr,"\n");
	     fprintf(stderr,"Error at offset %qd. Read %lu bytes.\n",
		     total_bytes,(long unsigned int)sizeof(buf1));
	     fprintf(stderr,"Files do not match.\n");
	     exit(3);
	 }
	 total_bytes += bytes1;
     }
     af_close(file1);
     af_close(file2);

     if(!opt_quiet){
	 printf("%60s\n","");
	 printf("Read %qd bytes. Files match!\n",total_bytes);
     }
}

void print_info(char dir,const char *segname,unsigned long arg,size_t len,
		unsigned char *data,int mcr)
{
    printf("    %c %s arg=%lu len=%d\n",dir,segname,arg,(int)len);
    printf("          ");
    if((arg == AF_SEG_QUADWORD) && (len==8)){
	printf("data=%"I64d" as a 64-bit value\n",af_decode_q(data));
	return;
    }
    /* Otherwise, just print some stuff... */
    for(unsigned int i=0;i<len && i<60;i++){
	if(data[i]==' '){
	    putchar(' ');
	    continue;
	}
	if(!isprint(data[i])){
	    putchar('.');
	    continue;
	}
	putchar(data[i]);
    }
    putchar('\n');
}

void compare_aff_metadata_segments(AFFILE *af1,AFFILE *af2,const char *segname,int mode)
{
    unsigned long arg1;
    size_t data1_len;
    int r1 = af_get_seg(af1,segname,&arg1,0,&data1_len);

    unsigned long arg2;
    size_t data2_len;
    int r2 = af_get_seg(af2,segname,&arg2,0,&data2_len);
    
    if(r1==0 && r2!=0){
	if(mode==1) printf("   %s  (arg=%lu; len=%d)\n",segname,arg1,(int)data1_len);
	return;
    }
    
    if(r2==0 && r1!=0){
	if(mode==2) printf("   %s  (arg=%lu; len=%d)\n",segname,arg2,(int)data2_len);
	return;
    }
    if(mode!=3) return;			// only report differences in mode 3
    /* Get the actual data... */
    unsigned char *data1 = (unsigned char *)malloc(data1_len);
    unsigned char *data2 = (unsigned char *)malloc(data2_len);

    r1 = af_get_seg(af1,segname,&arg1,data1,&data1_len);
    if(r1!=0) err(1,"Couldn't read data segment %s in %s",segname,af_filename(af1));

    r2 = af_get_seg(af2,segname,&arg2,data2,&data2_len);
    if(r2!=0) err(1,"Couldn't read data segment %s in %s",segname,af_filename(af2));

    int mcr = memcmp(data1,data2,data1_len);
    if(arg1!=arg2 || data1_len!=data2_len || mcr!=0){
	print_info('<',segname,arg1,data1_len,data1,mcr);
	print_info('>',segname,arg2,data2_len,data2,mcr);
	if(mcr){
	    printf("        *** Data segment are different ");
	    if(strcmp(segname,AF_BADFLAG)==0){
		printf("(but that's okay)");
	    }
	    putchar('\n');
	}
	putchar('\n');
    }
    free(data1);
    free(data2);
}

void compare_aff_data_segments(AFFILE *af1,AFFILE *af2,int segnum,int mode)
{
    char pagename[65];
    sprintf(pagename,AF_PAGE,segnum);

    char segname[65];
    sprintf(segname,AF_SEG_D,segnum);


    unsigned long arg1;
    size_t data1_len;
    int r1 = af_get_seg(af1,pagename,&arg1,0,&data1_len);
    if(r1==-1) r1=af_get_seg(af1,segname,&arg1,0,&data1_len);
    
    unsigned long arg2;
    size_t data2_len;
    int r2 = af_get_seg(af2,pagename,&arg2,0,&data2_len);
    if(r2 == -1) r2=af_get_seg(af2,segname,&arg2,0,&data2_len);
    
    if(r1<0 && r2<0) return;		// no data segment in either file
    if(r1==0 && r2!=0){
	if(mode==1) printf("   %s  (arg=%lu; len=%d)\n",segname,arg1,(int)data1_len);
	return;
    }
    
    if(r2==0 && r1!=0){
	if(mode==2) printf("   %s  (arg=%lu; len=%d)\n",segname,arg2,(int)data2_len);
	return;
    }
    if(mode!=3) return;			// only report differences in mode 3

    int pagenum = af_segname_page_number(segname);

    /* Get the actual data... */
    unsigned char *data1 = (unsigned char *)malloc(af_page_size(af1));
    unsigned char *data2 = (unsigned char *)malloc(af_page_size(af2));

    data1_len = af_page_size(af1);
    data2_len = af_page_size(af2);

    uint64 start_sector_number = (pagenum * data1_len) / 512;

    if(af_get_page(af1,pagenum,data1,&data1_len)<0)
	err(1,"Cannot read page %d from %s\n",pagenum,af_filename(af1));

    if(af_get_page(af2,pagenum,data2,&data2_len)<0)
	err(1,"Cannot read page %d from %s\n",pagenum,af_filename(af2));

    /* Now look at the pages sector-by-sector. */
    int af1_bad=0;
    int af2_bad=0;
    int matching_bad_blocks = 0;
    int matching_blocks = 0;
    int total_blocks = 0;
    int no_match = 0;
    vector<uint64> different_sectors;

    for(unsigned int offset=0;offset<data1_len;offset+=512){
	uint64 this_sector = start_sector_number + offset/512;
	total_blocks++;
	if(af_is_badblock(af1,data1+offset) &&
	   af_is_badblock(af2,data2+offset)){
	    matching_bad_blocks++;
	    continue;
	}
	if(af_is_badblock(af1,data1+offset)){
	    af1_bad++;
	    continue;
	}
	if(af_is_badblock(af2,data2+offset)){
	    af2_bad++;
	    continue;
	}
	if(memcmp(data1+offset,data2+offset,512)==0){
	    matching_blocks++;
	}
	else {
	    no_match++;
	    different_sectors.push_back(this_sector);
	}
    }

    printf("   page%d blocks:%d  matching: %d  no match: %d",
	   pagenum,total_blocks,matching_blocks,no_match);
    if(af1_bad) printf(" bad<: %d ",af1_bad);
    if(af2_bad) printf(" bad<: %d ",af2_bad);
    if(matching_bad_blocks) printf(" bad both: %d ",matching_bad_blocks);
    putchar('\n');
    if(opt_all && no_match>0){
	printf("  Sectors with differences:");
	int i=0;
	for(vector<uint64>::iterator j = different_sectors.begin();
	    j != different_sectors.end();
	    j++){
	    if(i==0){
		printf("\n   ");
	    }
	    printf(" %"I64d,*j);
	    i = (i+1) % 10;
	}
	putchar('\n');
    }
    if(opt_All){
	printf("here is the buffer\n");
	//fwrite(data1,1,af_page_size(af1),stdout);

	for(vector<uint64>::iterator j = different_sectors.begin();
	    j != different_sectors.end(); j++){
	    int offset = (*j - start_sector_number)*512;
	    char b2[16];
	    printf("offset=%d\n",offset);


	    memcpy(b2,data1+offset,16);
	    b2[15]=0;

	    printf("  sector %"I64d" ",*j);
	    printf("(offset=%d) first16:%s\n",offset,b2);
	    printf("   %s:\n",af_filename(af1));
	    print_sector(data1+offset);
	    printf("\n");
	    printf("   %s:\n",af_filename(af2));
	    //print_sector(data2+offset);
	    printf("\n\n");
	    //printf("bad1:\n"); print_sector(af1->badflag);
	    //printf("bad2:\n"); print_sector(af2->badflag);
	}
    }
    free(data1);
    free(data2);
}

void compare_aff_aff(const char *file1,const char *file2)
{
    bool no_data_segments = false;

    if(af_identify_file(file1)!=AF_IDENTIFY_AFF){
	errx(1,"%s must be an AFF file",file1);
    }
    if(af_identify_file(file2)!=AF_IDENTIFY_AFF){
	errx(1,"%s must be an AFF file",file2);
    }

    AFFILE *af1 = af_open(file1,O_RDONLY,0);
    AFFILE *af2 = af_open(file2,O_RDONLY,0);

    af_demand_sizes(af1);
    af_demand_sizes(af2);
    if(af1->image_pagesize != af2->image_pagesize){
	fprintf(stderr,"Currently, %s requires that both images have the "
		"same image datsegsize.\n"
		"pagesize(%s)=%ld\n"
		"pagesize(%s)=%ld\n",
		progname,file1,af1->image_pagesize,
		file2,af2->image_pagesize);
	fprintf(stderr,"Data segments will be ignored.\n");
	no_data_segments = true;
    }
    
    /* Compare all of the metadata segments in af1 with a2.
     * Report those that are missing or different. Then report
     * all of the segments in a2 but not in af1
     */

    /* First build a list of the segments in each */

    vector <string> segs_with_dups;

    printf("retrieving segment names...\n");

    AFFILE *af[2] = {af1,af2};

    for(int i=0;i<2;i++){
	af_rewind_seg(af[i]);
	char segname[AF_MAX_NAME_LEN];
	while(af_get_next_seg(af[i],segname,sizeof(segname),0,0,0)==0){
	    if(segname[0]){
		string s;
		s = segname;
		segs_with_dups.push_back(s);	// may give duplicates
	    }
	}
    }
    sort(segs_with_dups.begin(),segs_with_dups.end());
    vector<string>segs;

    /* Make a list of segs without duplicates */
    string last;
    for(vector<string>::iterator i = segs_with_dups.begin();i != segs_with_dups.end(); i++){
	if(last != *i){
	    segs.push_back(*i);
	}
	last = *i;
    }

    int lowest_segment = -1;
    int highest_segment = -1;
    /* Scan for the lowest and highest numbers */
    for(vector<string>::iterator i = segs.begin();i != segs.end(); i++){
	int num = af_segname_page_number(i->c_str());
	if(num!=-1){
	    if(num<lowest_segment  || lowest_segment==-1)  lowest_segment = num;
	    if(num>highest_segment || highest_segment==-1) highest_segment = num;
	}
    }


    if(opt_page != -1){
	lowest_segment = opt_page;
	highest_segment = opt_page;
    }


    if(opt_page == -1){
	puts("Inspecting metadata...");
	for(int mode=1;mode<=3;mode++){
	    switch(mode){
	    case 1: printf("  Segments only in %s:\n",af_filename(af1));break;
	    case 2: printf("  Segments only in %s:\n",af_filename(af2));break;
	    case 3: printf("  Segments in both files:\n");break;
	    }
	    
	    for(vector<string>::iterator i = segs.begin();i != segs.end(); i++){
		int num = af_segname_page_number(i->c_str());
		if(num==-1){
		    compare_aff_metadata_segments(af1,af2,i->c_str(),mode);
		}
	    }
	}
    }

    puts("Inspecting data...");
    for(int mode=1;mode<=3;mode++){
	switch(mode){
	case 1: printf("  Segments only in %s:\n",af_filename(af1));break;
	case 2: printf("  Segments only in %s:\n",af_filename(af2));break;
	case 3: printf("  Segments in both files:\n");break;
	}
	    
	for(int i=lowest_segment;i<=highest_segment;i++){
	    compare_aff_data_segments(af1,af2,i,mode);
	}
    }

}

int main(int argc,char **argv)
{
    int bflag, ch;

     progname = argv[0];
     bflag = 0;
     while ((ch = getopt(argc, argv, "s:h?rqb:aAn:p:")) != -1) {
	 switch (ch) {
	 case 'r':
	     opt_remove++;
	     break;
	 case 'q':
	     opt_quiet++;
	     break;
	 case 'b':
	     opt_batch++;
	     batch_ext=optarg;
	     break;
	 case 'a':
	     opt_all++;
	     break;
	 case 'A':
	     opt_All++;
	     break;
	 case 'p':
	     opt_page = atoi(optarg);
	     break;
	 case 'h':
	 case '?':
	 default:
	     usage();
	 }
     }
     argc -= optind;
     argv += optind;

     while(opt_batch && argc){
	 char *aff_file = *argv++;
	 argc--;
	 char *iso_file = (char *)malloc(strlen(aff_file)+16);
	 char *cc;

	 strcpy(iso_file,aff_file);
	 cc = rindex(iso_file,'.');
	 if(cc){
	     strcpy(cc+1,batch_ext);
	 }
	 else {
	     strcat(iso_file,".");
	     strcat(iso_file,batch_ext);
	 }
	 compare_aff_iso(aff_file,iso_file);
	 free(iso_file);
	 if(argc==0) exit(0);		// did them all
     }

     if(argc==2){			// if just 2, compare them
	 char *file1 = *argv++;
	 char *file2 = *argv++;
	 
	 if(af_identify_file(file1)==1 && af_identify_file(file2)==1){
	     compare_aff_aff(file1,file2);
	 }
	 else{
	     compare_aff_iso(file1,file2);
	 }
	 exit(0);
     }
     usage();				// don't know what to do
     exit(0);
}
